/*
 *  ADP (Another Data Processor) www.adp.la
 *  Copyright (C) 2010 Katsuhisa Ohfuji <katsuhisa@ohfuji.name>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */

#ifndef ADP_BUILTIN_STRING_H
#define ADP_BUILTIN_STRING_H

struct ExecContext_Regex_Search : public ExecContextBase {
	// regex
	boost::regex			reg;
	const string			*text;
	string::const_iterator	start;

	ExecContext_Regex_Search(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextBase(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 2 ) return RERR_argment_number(excp, 2);

		string src, find;
		const string *psrc, *pfind;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);

		string temp,regexnocrlf;

		replace_all( *pfind, "\n", "", temp);
		replace_all( temp, "\r", "", regexnocrlf);

		try {
			reg = boost::regex( regexnocrlf.c_str() );
		} catch(...) {
			return RERR_regex_format_error(excp);
		}
		text = psrc;
		start = text->begin();
		return backtrack(excp);
	}
	virtual bool backtrack(PException &excp) {
		boost::smatch what;
		try {
			bool ret = boost::regex_search( start, static_cast<string::const_iterator>(text->end()), what, reg );
			if ( ret ) {
				PArray *alloc_ary = 0;
				if ( !size_check_and_alloc_array( what.size() - 1 + 2, excp, &alloc_ary) ) return false;
				start = what[0].second;
				size_t	vidx = 2;
				for ( size_t i = 1; i < what.size(); i++ ) {
					PString *po = pmm.newPString(pobjs, string(what[i].first, what[i].second));
					if ( !unify<PString>( vidx, po, excp) ) return false;
					vidx++;
				}
			}
			return ret;
		} catch(...) {
			return RERR_regex_format_error(excp);
		}
	}
};


struct ExecContext_Cat : public ExecContextRoot {
	ExecContext_Cat(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 2 ) return RERR_argment_number(excp, 2);
		PString  *result = pmm.newPString(pobjs);
		size_t last = pred->size() - 1;
		for ( size_t i = 0; i < last; i++ ) {
			string src;
			const string *psrc;
			if ( (psrc = cnv_string( i, src)) == 0 ) {
				return RERR_argment_type(excp, i);
			}
			result->value += *psrc;
		}
		return unify<PString>( last, result, excp);
	}
};

struct ExecContext_Left : public ExecContextRoot {
	ExecContext_Left(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 3 ) return RERR_argment_number(excp, 3);

		string src;
		const string *psrc;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		bool r;
		PINTEGER s = cnv_integer( 1, r);
		if ( !r ) return RERR_argment_type(excp, 1);

		if ( s < 0 ) return false;

		PString  *result = pmm.newPString(pobjs);
		result->value = psrc->substr( 0, static_cast<size_t>(s));
		return unify<PString>( 2, result, excp);
	}
};


struct ExecContext_Right : public ExecContextRoot {
	ExecContext_Right(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 3 ) return RERR_argment_number(excp, 3);

		string src;
		const string *psrc;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		bool r;
		PINTEGER s = cnv_integer( 1, r);
		if ( !r ) return RERR_argment_type(excp, 1);

		if ( s < 0 ) return false;

		PString  *result = pmm.newPString(pobjs);
		result->value = psrc->substr( psrc->size() - static_cast<size_t>(s), static_cast<size_t>(s));
		return unify<PString>( 2, result, excp);
	}
};

struct ExecContext_Mid : public ExecContextRoot {
	ExecContext_Mid(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 4 ) return RERR_argment_number(excp, 4);

		string src;
		const string *psrc;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		bool r;
		PINTEGER b = cnv_integer( 1, r);
		if ( !r ) return RERR_argment_type(excp, 1);
		PINTEGER s = cnv_integer( 2, r);
		if ( !r ) return RERR_argment_type(excp, 2);

		if ( b < 0 || s < 0 || s > psrc->size() ) return false;

		PString  *result = pmm.newPString(pobjs);
		result->value = psrc->substr( static_cast<size_t>(b), static_cast<size_t>(s));
		return unify<PString>( 3, result, excp);
	}
};

struct ExecContext_Chr : public ExecContextRoot {
	ExecContext_Chr(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);
		bool r;
		PINTEGER v = cnv_integer( 0, r);
		if ( !r ) {
			const PNil *p = get<PNil>(0);
			if ( p ) {
				return (*pred)[1]->unify( *p, this);
			} else {
				return RERR_argment_type(excp, 0);
			}
		}

		PString  *result = pmm.newPString(pobjs);
		result->value.push_back( static_cast<char>(v) );
		return unify<PString>( 1, result, excp);
	}
};

struct ExecContext_Ascii : public ExecContextRoot {
	ExecContext_Ascii(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);
		const PNil *p = get<PNil>(0);
		if ( p ) {
			return (*pred)[1]->unify( *p, this);
		} else {
			string src;
			const string *psrc;
			if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
			if ( psrc->empty() ) {
				return (*pred)[1]->unify( pnil, this);
			} else {
				const PInteger  *result = pmm.newPInteger(pobjs, (*psrc)[0]);
				return unify<PInteger>( 1, result, excp);
			}
		}
	}
};

struct ExecContext_Instr : public ExecContextBase {
	const string *psrc;
	const string *pfind;
	size_t pos;

	ExecContext_Instr(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextBase(p_, f_, pred_, l), psrc(0) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 3 && pred->size() != 4) return RERR_argment_number(excp, 3);

		string src, find;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);

		if ( pred->size() == 3 ) {
			pos = -1;
		} else {
			bool r;
			pos = static_cast<size_t>(cnv_integer( 2, r) - 1);	// ̂Ƀ[vׂ+1Ă̂-1Ă 
			if ( !r ) return RERR_argment_type(excp, 2);
		}
		return backtrack(excp);
	}

	virtual bool backtrack(PException &excp) {
		pos = psrc->find( *pfind, pos + 1);
		if ( pos == string::npos ) return false;
		const PInteger  *result = pmm.newPInteger(pobjs, pos);
		return unify<PInteger>( pred->size() - 1, result, excp);
	}
};


struct ExecContext_InstrRev : public ExecContextBase {
	const string *psrc;
	const string *pfind;
	size_t rpos;

	ExecContext_InstrRev(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextBase(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 3 && pred->size() != 4) return RERR_argment_number(excp, 3);

		string src, find;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);

		if (  pred->size() == 3 ) {
			rpos = psrc->size();
		} else {
			bool r;
			rpos = static_cast<size_t>(cnv_integer( 2, r) + 1);	// ̂Ƀ[vׂ-1Ă̂+1Ă 
			if ( !r ) return RERR_argment_type(excp, 2);
		}
		return backtrack(excp);
	}

	virtual bool backtrack(PException &excp) {
		while ( true ) {
			rpos = psrc->find_last_of( *pfind, static_cast<size_t>(rpos - 1));
			if ( rpos == string::npos ) return false;
			if ( memcmp( psrc->data() + rpos, pfind->data(), pfind->size()) == 0 ) {
				break;
			}
			if ( rpos == 0 ) return false;
		}
		const PInteger  *result = pmm.newPInteger(pobjs, rpos);
		return unify<PInteger>( pred->size() - 1, result, excp);
	}
};


struct ExecContext_Replace : public ExecContextRoot {
	ExecContext_Replace(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 4 || 5 < pred->size() ) return RERR_argment_number(excp, 4);

		string src, find, replace;
		const string *psrc, *pfind, *preplace;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);
		if ( (preplace = cnv_string( 2, replace)) == 0 ) return RERR_argment_type(excp, 2);
		size_t off = 0;
		if ( pred->size() == 5 ) {
			bool r;
			off = static_cast<size_t>(cnv_integer(3, r));
			if ( !r ) return RERR_argment_type(excp, 3);
		}

		PString *result = pmm.newPString( pobjs);
		replace_all( *psrc, *pfind, *preplace, result->value, off);
		return unify<PString>( pred->size() - 1, result, excp);
	}
};

struct ExecContext_ReplaceFirst : public ExecContextRoot {
	ExecContext_ReplaceFirst(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 4 || 5 < pred->size() ) return RERR_argment_number(excp, 4);
		
		string src, find, replace;
		const string *psrc, *pfind, *preplace;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);
		if ( (preplace = cnv_string( 2, replace)) == 0 ) return RERR_argment_type(excp, 2);
		size_t off = 0;
		if ( pred->size() == 5 ) {
			bool r;
			off = static_cast<size_t>(cnv_integer(3, r));
			if ( !r ) return RERR_argment_type(excp, 3);
		}

		PString *result = pmm.newPString( pobjs);
		replace_first( *psrc, *pfind, *preplace, result->value, off);
		return unify<PString>( pred->size() - 1, result, excp);
	}
};

struct ExecContext_RegexReplace : public ExecContextRoot {
	ExecContext_RegexReplace(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 4 || 5 < pred->size() ) return RERR_argment_number(excp, 4);
		string src, find, replace;
		const string *psrc, *pfind, *preplace;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);
		if ( (preplace = cnv_string( 2, replace)) == 0 ) return RERR_argment_type(excp, 2);
		size_t off = 0;
		if ( pred->size() == 5 ) {
			bool r;
			off = static_cast<size_t>(cnv_integer(3, r));
			if ( !r ) return RERR_argment_type(excp, 3);
		}

		PString *result = pmm.newPString( pobjs);
		replace_regex_all( *psrc, pfind->c_str(), preplace->c_str(), result->value, off);
		return unify<PString>( pred->size() - 1, result, excp);
	}
};

struct ExecContext_RegexReplaceFirst : public ExecContextRoot {
	ExecContext_RegexReplaceFirst(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 4 || 5 < pred->size() ) return RERR_argment_number(excp, 4);

		string src, find, replace;
		const string *psrc, *pfind, *preplace;
		if ( (psrc = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (pfind = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);
		if ( (preplace = cnv_string( 2, replace)) == 0 ) return RERR_argment_type(excp, 2);
		size_t off = 0;
		if ( pred->size() == 5 ) {
			bool r;
			off = static_cast<size_t>(cnv_integer(3, r));
			if ( !r ) return RERR_argment_type(excp, 3);
		}

		PString *result = pmm.newPString( pobjs);
		replace_regex_first( *psrc, pfind->c_str(), preplace->c_str(), result->value, off);
		return unify<PString>( pred->size() - 1, result, excp);
	}
};


struct ExecContext_Trim : public ExecContextRoot {
	ExecContext_Trim(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 2 ) return RERR_argment_number(excp, 2);
		string src;
		const string *psrc = cnv_string( 0, src);
		if ( !psrc ) return RERR_argment_type(excp, 0);

		string::const_iterator l = psrc->begin();
		string::const_iterator r = psrc->end() - 1;

		/* LTrim */
		while ( *l == ' ' || *l == '\t' || *l == '\r' || *l == '\n' ) {
			if ( l < psrc->end() ) 
				l++;
			else
				break;
			
		}
		/* RTrim */
		while ( *r == ' ' || *r == '\t' || *r == '\r' || *r == '\n' ) {
			if ( r > psrc->begin() ) 
				r--;
			else
				break;
		}

		PString *result = pmm.newPString( pobjs);
		result->value.assign(l,r+1);
		return unify<PString>( 1, result, excp);
	}
};

struct ExecContext_Sprintf : public ExecContextRoot {
	ExecContext_Sprintf(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 2 ) return RERR_argment_number(excp, 2);
		const char	*f = get_charp(0);
		if ( !f ) return RERR_argment_type(excp, 0);
		PString *p = pmm.newPString( pobjs);
		string &out = p->value;

		size_t	len = strlen(f);
		size_t	idx = 1;
		size_t	i = 0;	
		while ( i < len ) {
			if ( f[i] == '%' ) {
				i++;
				string	flags;	// -+0 #
				size_t	width = (size_t)-1;
				size_t	precision = (size_t)-1;
				char	type = 0;	// s,d
				// get flgs
				while ( strchr("-+0 #", f[i]) != 0 ) {
					flags.push_back(f[i]);
					i++;
				}
				// get width;
				string buf;
				while ( isdigit(f[i]) ) {
					buf.push_back(f[i]);
					i++;
				}
				if ( !buf.empty() ) width = atoi(buf.c_str());
				buf.clear();
				// get precision
				if ( f[i] == '.' ) {
					i++;
					while ( isdigit(f[i]) ) {
						buf.push_back(f[i]);
						i++;
					}
					if ( !buf.empty() ) precision = atoi(buf.c_str());
				}
				// type
				switch ( f[i] ) {
					case 's':
						{
							string str;
							const char *sp = cnv_charp(idx++, str);
							if ( !sp ) return  RERR_argment_type(excp, idx-1);
							size_t	slen = strlen(sp);
							if ( width == (size_t)-1 ) width = slen;
							if ( precision == (size_t)-1) precision = slen;
							bool leftflg = (strchr( flags.c_str(), '-') != 0);
							if ( width > slen && !leftflg ) {
								out += string( width - slen, ' ');
							}
							if ( precision < slen ) {
								out += string( sp, sp + precision);
							} else {
								out += sp;
							}
							if ( width > slen && leftflg ) {
								out += string( width - slen, ' ');
							}
							i++;
						}
						break;
					case 'd':
						{
							const PObject *pobj = get<PObject>(idx++);
							if ( !pobj ) return RERR_argment_type(excp, idx-1);
							PINTEGER d;
							if ( !pobj->cnv_integer(d) ) return RERR_argment_type(excp, idx-1);
							char	fmt[64];
							char	buf[64];
							if ( ( width == (size_t) - 1 ) && ( precision == (size_t) - 1 ) ) {
								sprintf( fmt, "%%%slld", flags.c_str());
							} else if ( width == (size_t) - 1 ) {
								sprintf( fmt, "%%%s.%lldlld", flags.c_str(), (PINTEGER)precision);
							} else if ( precision == (size_t) - 1 ) {
								sprintf( fmt, "%%%s%lldlld", flags.c_str(), (PINTEGER)width);
							} else {
								sprintf( fmt, "%%%s%lld.%lldlld", flags.c_str(), (PINTEGER)width, (PINTEGER)precision);
							}
							sprintf( buf, fmt, d);
							out += buf;
							i++;
						}
						break;
					case 'x':
						{
							const PObject *pobj = get<PObject>(idx++);
							if ( !pobj ) return RERR_argment_type(excp, idx-1);
							PINTEGER	d;
							if ( !pobj->cnv_integer(d) ) return RERR_argment_type(excp, idx-1);
							char	fmt[64];
							char	buf[64];
							if ( ( width == (size_t) - 1 ) && ( precision == (size_t) - 1 ) ) {
								sprintf( fmt, "%%%sllx", flags.c_str());
							} else if ( width == (size_t) - 1 ) {
								sprintf( fmt, "%%%s.%lldllx", flags.c_str(), (PINTEGER)precision);
							} else if ( precision == (size_t) - 1 ) {
								sprintf( fmt, "%%%s%lldllx", flags.c_str(), (PINTEGER)width);
							} else {
								sprintf( fmt, "%%%s%lld.%lldllx", flags.c_str(), (PINTEGER)width, (PINTEGER)precision);
							}
							sprintf( buf, fmt, d);
							out += buf;
							i++;
						}
						break;
					case 'o':
						{
							const PObject *pobj = get<PObject>(idx++);
							if ( !pobj ) return RERR_argment_type(excp, idx-1);
							PINTEGER	d;
							if ( !pobj->cnv_integer(d) ) return RERR_argment_type(excp, idx-1);
							char	fmt[64];
							char	buf[64];
							if ( ( width == (size_t) - 1 ) && ( precision == (size_t) - 1 ) ) {
								sprintf( fmt, "%%%sllo", flags.c_str());
							} else if ( width == (size_t) - 1 ) {
								sprintf( fmt, "%%%s.%lldllo", flags.c_str(), (PINTEGER)precision);
							} else if ( precision == (size_t) - 1 ) {
								sprintf( fmt, "%%%s%lldllo", flags.c_str(), (PINTEGER)width);
							} else {
								sprintf( fmt, "%%%s%lld.%lldllo", flags.c_str(), (PINTEGER)width, (PINTEGER)precision);
							}
							sprintf( buf, fmt, d);
							out += buf;
							i++;
						}
						break;
					case 'f':
						{
							const PObject *pobj = get<PObject>(idx++);
							if ( !pobj ) return RERR_argment_type(excp, idx-1);
							double	d;
							if ( !pobj->cnv_double(d) ) return RERR_argment_type(excp, idx-1);
							char	fmt[64];
							char	buf[64];
							if ( ( width == (size_t) - 1 ) && ( precision == (size_t) - 1 ) ) {
								sprintf( fmt, "%%%sf", flags.c_str());
							} else if ( width == (size_t) - 1 ) {
								sprintf( fmt, "%%%s.%lldf", flags.c_str(), (PINTEGER)precision);
							} else if ( precision == (size_t) - 1 ) {
								sprintf( fmt, "%%%s%lldf", flags.c_str(), (PINTEGER)width);
							} else {
								sprintf( fmt, "%%%s%lld.%lldf", flags.c_str(), (PINTEGER)width, (PINTEGER)precision);
							}
							sprintf( buf, fmt, d);
							out += buf;
							i++;
						}
						break;
					case ',':
						{
							const PObject *pobj = get<PObject>(idx++);
							if ( !pobj ) return RERR_argment_type(excp, idx-1);
							PINTEGER	d;
							if ( !pobj->cnv_integer(d) ) return RERR_argment_type(excp, idx-1);
							buf.clear();
							formatNumber(d, buf);
							out += buf;
							i++;
						}
						break;
					case '%':
						out.push_back(f[i]);
						i++;
				}
			} else {
				out.push_back(f[i]);
				i++;
			}
		}
		return unify<PString>( idx, p, excp);
	}
};

struct ExecContext_Split : public ExecContextRoot {
	ExecContext_Split(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 3 ) return RERR_argment_number(excp, 3);

		string src, find;
		const string *s, *f;
		if ( (s = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		if ( (f = cnv_string( 1, find)) == 0 ) return RERR_argment_type(excp, 1);

		size_t			off = 0;
		size_t			flen = f->size();
		size_t			fnd;
		vector<size_t>	spos;
		spos.push_back(0);
		while ( (fnd = s->find( *f, off)) != string::npos ) {
			spos.push_back(fnd);
 			off = fnd + flen;
			spos.push_back(off);
		}
		spos.push_back(s->size());

		PObject	*list = &pnil;
		for ( vector<size_t>::reverse_iterator i = spos.rbegin(); i < spos.rend(); i++ ) {
			size_t	end  = *i++;
			size_t	begin = *i;
			PString	*po = pmm.newPString(pobjs, s->begin() + begin, s->begin() + end);
			PList *nlist = pmm.newPList(pobjs, po, list, true);
			list = nlist;
		}
		return unify<PObject>(2, list, excp);
	}
};

struct ExecContext_Euc2sjis : public ExecContextRoot {
	ExecContext_Euc2sjis(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);

		string src;
		const string *s;
		if ( (s = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		
		PString *result = pmm.newPString( pobjs);

		result->value.reserve(s->capacity());
		char sj[3];
		size_t siz = s->size();

		for ( size_t i = 0; i < siz; i++ ) {
			if ( (*s)[i] < 0x80 ) {	// `FbN܂
				result->value += (*s)[i];
			} else if ( i + 1 < siz ) {
				ej2sj( (const unsigned char*)(s->data()+i), (unsigned char *)sj);
				i++;
				result->value += (char*)sj;
			}
		}
		return unify<PString>( 1, result, excp);
	}
};

struct ExecContext_FromUTF8 : public ExecContextRoot {
	ExecContext_FromUTF8(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);

		string src;
		const string *s;
		if ( (s = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);

		PString *result = pmm.newPString(pobjs);
		if ( utf8to( *s, result->value) == false ) return  RERR_setlocale(excp);
		return unify<PString>( 1, result, excp);
	}
};


struct ExecContext_CSV : public ExecContextRoot {
	ExecContext_CSV(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() < 2 ) return RERR_argment_number(excp, 2);
		PString *result = pmm.newPString( pobjs);

		size_t last = pred->size() - 1;
		result->value.reserve(1024);
		for ( size_t i = 0; i < last; i++ ) {
			if ( i != 0 ) { result->value += ','; }
			const PObject *item = get<PObject>(i);
			item->cnv_csv(result->value, ',', '"');
		}
		return unify<PString>( last, result, excp);
	}
};

struct ExecContext_JSON : public ExecContextRoot {
	ExecContext_JSON(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);
		PString *result = pmm.newPString( pobjs);
		result->value.reserve(1024);
		get<PObject>(0)->cnv_json(result->value);
		return unify<PString>( 1, result, excp);
	}
};

struct ExecContext_ParseJSON : public ExecContextRoot {
	ExecContext_ParseJSON(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);

		string src;
		const string *s;
		if ( (s = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);
		size_t pos = 0;

		PObject *po = get_gc()->factory.createPObjectJSON( *s, pos, pobjs);
		if ( po != 0 ) { po = po->parseJson( *s, pos, pobjs); }
		if ( po == 0 ) return RERR_Convert( excp, pos);

		return unify<PObject>( 1, po, excp);
	}
};

struct ExecContext_SHA1 : public ExecContextRoot {
	ExecContext_SHA1(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);

		string src;
		const string *s;
		if ( (s = cnv_string( 0, src)) == 0 ) return RERR_argment_type(excp, 0);

		PString *result = pmm.newPString( pobjs);

		if ( !createHash( *s, result->value) ) return RERR_System_Call(excp);
	
		return unify<PString>( 1, result, excp);
	}
};

struct ExecContext_MakeSessionID : public ExecContextRoot {
	ExecContext_MakeSessionID(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
		if ( pred->size() != 2 ) return RERR_argment_number(excp, 2);
		PString *t = pmm.newPString( pobjs);
		PString *p = pmm.newPString( pobjs);

		if ( !createSessionID( p->value, t->value, envmap) ) return RERR_File_Operation(excp);

		return unify<PString>( 0, t, excp) && unify<PString>( 1, p, excp);
	}
};

struct ExecContext_CTimeStr : public ExecContextRoot {
	ExecContext_CTimeStr(ExecContext *p_, void *f_, const PPredicate *pred_, VLocal *l) : ExecContextRoot(p_, f_, pred_, l) {}
	virtual bool first(PException &excp) { 
static const char *wtbl[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
static const char *mtbl[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
		if ( pred->size() != 4 ) return RERR_argment_number(excp, 4);
		const PString *psrc = get<PString>(0);
		const PString *pfmt = get<PString>(1);
		const PString *pomt = get<PString>(2);
		if ( !psrc ) return RERR_argment_type( excp, 0);
		if ( !pfmt ) return RERR_argment_type( excp, 1);
		if ( !pomt ) return RERR_argment_type( excp, 2);
		const string &src = psrc->value;
		const string &fmt = pfmt->value;
		const string &omt = pomt->value;

		int year = 0;
		int month = 0;
		int day = 0;
		int hour = 0;
		int minute = 0;
		int secound = 0;
		int week = 0;
		//int timezone = 0;

		// DateTime̎擾
		size_t spos = 0;
		size_t fpos = 0;
		while ( spos < src.size() && fpos < fmt.size() ) {
			if ( fmt[fpos] == '%' ) {
				int i;
				fpos++;
				if ( fpos >= fmt.size() ) return RERR(excp, "format error(%) ");
				switch( fmt[fpos] ) {
					case 'a' : 
						for ( i = 0; i < sizeof(wtbl)/sizeof(wtbl[0]); i++ ) {
							if ( memcmp( src.c_str() + spos, wtbl[i], 3) == 0 ) break;
						}
						if ( i >= sizeof(wtbl)/sizeof(wtbl[0]) ) return RERR(excp," format error(a) ");
						week = i;
						spos += 3;
						fpos++;
						break;
					case 'b' : 
						for ( i = 0; i < sizeof(mtbl)/sizeof(mtbl[0]); i++ ) {
							if ( memcmp( src.c_str() + spos, mtbl[i], 3) == 0 ) break;
						}
						if ( i >= sizeof(mtbl)/sizeof(mtbl[0]) ) return RERR(excp," format error(b) ");
						month = i + 1; // 1n܂
						spos += 3;
						fpos++;
						break;
					case 'd' :
						if ( sscanf( src.c_str() + spos, "%2d", &day) != 1 ) return RERR(excp," format error(d) ");
						if ( day < 1) {
							return false;
						} else if ( day <= 31 ) {
							spos +=2;
						} else {
							return false;
						}
						fpos++;
						break;
					case 'H' :
						if ( sscanf( src.c_str() + spos, "%2d", &hour) != 1 ) return RERR(excp," format error(H) ");
						if ( hour < 0) {
							return false;
						} else if ( hour <= 23 ) {
							spos +=2;
						} else {
							return false;
						}
						fpos++;
						break;
					case 'm' :
						if ( sscanf( src.c_str() + spos, "%2d", &month) != 1 ) return RERR(excp," format error(m) ");
						if ( month < 1 ) {
							return false;
						} else if ( month <= 12 ) {
							spos +=2;
						} else {
							return false;
						}
						fpos++;
						break;
					case 'M' :
						if ( sscanf( src.c_str() + spos, "%2d", &minute) != 1 ) return RERR(excp," format error(M) ");
						if ( minute < 0 ) {
							return false;
						} else if ( minute <= 59 ) {
							spos +=2;
						} else {
							return false;
						}
						fpos++;
						break;
					case 'S' :
						if ( sscanf( src.c_str() + spos, "%2d", &secound) != 1 ) return RERR(excp," format error(S) ");
						if ( secound < 0) {
							return false;
						} else if ( secound <= 60 ) {
							spos +=2;
						} else {
							return false;
						}
						fpos++;
						break;
					case 'w' :
						if ( sscanf( src.c_str() + spos, "%1d", &week) != 1 ) return RERR(excp," format error(w) ");
						if ( week < 0) {
							return false;
						} else if ( week <= 6 ) {
							spos +=1;
						} else {
							return false;
						}
						fpos++;
						break;
					case 'Y' :
						if ( sscanf( src.c_str() + spos, "%4d", &year) != 1 ) return RERR(excp," format error(Y) ");
						if ( year < 0 ) {
							return false;
						} else if ( year < 10000 ) {
							spos +=4;
						} else {
							return false;
						}
						fpos++;
						break;
					case '%' :
						break; // % LƂĔr
					default :
						return RERR_argment_type( excp, 1); // tH[}bgG[
				}
			} else {
				if ( src[spos] != fmt[fpos] ) return RERR(excp," format error ");
				spos++;
				fpos++;
			}
		}
		if ( fpos < fmt.size() ) return false;
		
		
		// DateTimȅo
		PString *p = pmm.newPString( pobjs);
		string &buf = p->value;

		size_t opos = 0;
		while ( opos < omt.size() ) {
			if ( omt[opos] == '%' ) {
				opos++;
				if ( opos >= omt.size() ) break;
				switch( omt[opos] ) {
				case 'a' :
					buf += wtbl[week];
					opos++;
					break;
				case 'b' :
					buf += mtbl[month-1];
					opos++;
					break;
				case 'd' :
					buf += cformat("%02d", day);
					opos++;
					break;
				case 'H' :
					buf += cformat("%02d", hour);
					opos++;
					break;
				case 'm' :
					buf += cformat("%02d", month);
					opos++;
					break;
				case 'M' :
					buf += cformat("%02d", minute);
					opos++;
					break;
				case 'S' :
					buf += cformat("%02d", secound);
					opos++;
					break;
				case 'w' :
					buf += cformat("%01d", week);
					opos++;
					break;
				case 'Y' :
					buf += cformat("%04d", year);
					opos++;
					break;
				default :
					return RERR_argment_type( excp, 2);
				}
			} else {
				buf.push_back(omt[opos]);
				opos++;
			}
		}
		return unify<PString>( 3, p, excp);
	}
};

#endif
